package org.expeditee.gio.javafx;

import java.util.HashSet;
import java.util.Set;
import java.util.Stack;

import org.expeditee.Util;
import org.expeditee.core.Clip;
import org.expeditee.core.Colour;
import org.expeditee.core.Cursor;
import org.expeditee.core.Dimension;
import org.expeditee.core.Fill;
import org.expeditee.core.Font;
import org.expeditee.core.GradientFill;
import org.expeditee.core.Image;
import org.expeditee.core.Point;
import org.expeditee.core.Stroke;
import org.expeditee.core.TextLayout;
import org.expeditee.core.bounds.AxisAlignedBoxBounds;
import org.expeditee.core.bounds.PolygonBounds;
import org.expeditee.gio.GraphicsManager;
import org.expeditee.gio.GraphicsSurfaceStack;

import javafx.application.Application;
import javafx.application.Platform;
import javafx.geometry.Bounds;
import javafx.scene.Group;
import javafx.scene.Scene;
import javafx.scene.SnapshotParameters;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.effect.BlendMode;
import javafx.scene.input.KeyCombination;
import javafx.scene.paint.Color;
import javafx.stage.Screen;
import javafx.stage.Stage;

public class JavaFXGraphicsManager extends GraphicsManager {
	
	/** Singleton instance. */
	private static JavaFXGraphicsManager _instance;
	
	/** Singleton instantiator. */
	public static JavaFXGraphicsManager getInstance()
	{
		if (_instance == null) _instance = new JavaFXGraphicsManager();
		
		return _instance;
	}

	/** The largest size a window icon can be. */
	private static final int MAX_ICON_SIZE = 256;
	/** The smallest size a window icon can be. */
	private static final int MIN_ICON_SIZE = 8;
	
	/** The startup thread for the JavaFX application. */
	private static Thread _jfxAppThread;
	
	/** The stack of drawing surfaces. */
	private GraphicsSurfaceStack<Canvas> _surfaceStack;
	
	/** A stack for surface clips to aid in clip clearing. */
	private Clip _currentClip;
	private Stack<Clip> _clipStack;
	
	private JavaFXGraphicsManager()
	{
		// Create the clip stack
		_currentClip = null;
		_clipStack = new Stack<Clip>();
		
		// Set the ExpediteeApplication manager
		ExpediteeApplication.manager = this;
		
		// Start up the JavaFX thread on another thread as it blocks until program exit
		_jfxAppThread = new Thread("_jfxAppThread") {
			@Override
			public void run()
			{
				Application.launch(ExpediteeApplication.class);
			}
		};
		
		// Block the main thread until the JavaFX thread is ready
		synchronized(_jfxAppThread) {
			_jfxAppThread.start();
			try {
				_jfxAppThread.wait();
			} catch (InterruptedException e) {
				System.err.println("Error while initialising GraphicsManager. Aborting...");
				e.printStackTrace(System.err);
				System.exit(1);
			}
		}
		
		// Create the surface stack
		_surfaceStack = new GraphicsSurfaceStack<Canvas>() {
			@Override
			public Canvas getSurfaceFromImage(Image image)
			{
				Canvas canvas = new Canvas(image.getWidth(), image.getHeight());
				canvas.getGraphicsContext2D().drawImage(JavaFXMiscManager.getIfUsingJavaFXImageManager().getInternalImage(image), 0, 0);
				return canvas;
			}
			
			@Override
			public void setSurfaceClip(Canvas surface, Clip clip)
			{
				if (surface == null) return;
				
				if (clip != null && clip.isFullyClipped()) return;
				
				GraphicsContext g = surface.getGraphicsContext2D();
				
				// Restore the pre-clipped state
				g.restore();
				
				// If we are clearing the clip, we are done
				if (clip == null || clip.getBounds() == null) return;
				
				// Re-save the pre-clipped state
				g.save();
				
				// Set the clip
				g.beginPath();
				AxisAlignedBoxBounds clipBounds = clip.getBounds();
				g.rect(clipBounds.getMinX(), clipBounds.getMinY(), clipBounds.getWidth(), clipBounds.getHeight());
				g.closePath();
				g.clip();
			}
		};
		_surfaceStack.setRootSurface(ExpediteeApplication.theCanvas);
		
		// Disable JavaFX's fullscreen exiting system
		ExpediteeApplication.theStage.setFullScreenExitHint("");
		ExpediteeApplication.theStage.setFullScreenExitKeyCombination(KeyCombination.NO_MATCH);
		
		// Block until the window is ready
		JavaFXMiscManager.waitUntilPropertyEquals(ExpediteeApplication.theStage.showingProperty(), true);
		
	}
	
	@Override
	protected GraphicsSurfaceStack<?> getGraphicsSurfaceStack()
	{
		return _surfaceStack;
	}

	@Override
	public Dimension getScreenSize()
	{
		return JavaFXConversions.fromJavaFXRectangle2D(Screen.getPrimary().getBounds()).getSize();
	}

	@Override
	public Point getWindowLocation()
	{
		return translateCanvasToScreen(new Point(0, 0));
	}

	@Override
	public void setWindowLocation(Point p)
	{
		if (p == null) return;
		
		Point currentCanvasPos = getWindowLocation();
		Point currentStagePos = new Point((int) ExpediteeApplication.theStage.getX(), (int) ExpediteeApplication.theStage.getY());
		
		Point delta = Point.difference(currentStagePos, currentCanvasPos);
		
		Point newStagePos = p.clone().add(delta);
		
		ExpediteeApplication.theStage.setX(newStagePos.getX());
		ExpediteeApplication.theStage.setY(newStagePos.getY());
	}

	@Override
	public Dimension getWindowSize()
	{
		return new Dimension((int) ExpediteeApplication.theScene.getWidth(), (int) ExpediteeApplication.theScene.getHeight());
	}

	@Override
	public void setWindowSize(Dimension d)
	{
		if (d == null) return;

		Dimension decorationDimensions = getDecorationDimensions();
		
		ExpediteeApplication.theStage.setWidth(d.width + decorationDimensions.width);
		ExpediteeApplication.theStage.setHeight(d.height + decorationDimensions.height);
		ExpediteeApplication.theCanvas.setWidth(d.width);
		ExpediteeApplication.theCanvas.setHeight(d.height);
		
	}

	@Override
	public void requestFocus()
	{
		ExpediteeApplication.theStage.requestFocus();
	}

	@Override
	public void setWindowIcon(Image image)
	{
		// TODO: Work out why this doesn't work. cts16
		JavaFXImageManager imageManager = JavaFXMiscManager.getIfUsingJavaFXImageManager();
		
		if (imageManager == null || !imageManager.isImageValid(image)) return;

		ExpediteeApplication.theStage.getIcons().clear();

		ExpediteeApplication.theStage.getIcons().add(imageManager.getInternalImage(image));
		
		/*
		Set<Image> iconSet = createIconSet(image);
		
		if (iconSet != null) {
			for (Image icon : iconSet) ExpediteeApplication.theStage.getIcons().add(imageManager.getInternalImage(icon));
		}
		*/
	}
	
	@Override
	public void setWindowTitle(String title)
	{
		if (title == null) return;
		
		ExpediteeApplication.theStage.setTitle(title);
	}
	
	private Point _preFullscreenWindowLocation = null;
	private Dimension _preFullscreenWindowSize = null;

	@Override
	public boolean canGoFullscreen()
	{
		// TODO: Check if this is always true. cts16
		return true;
	}

	@Override
	public boolean isFullscreen()
	{
		return ExpediteeApplication.theStage.isFullScreen();
	}

	@Override
	public void goFullscreen()
	{
		_preFullscreenWindowLocation = getWindowLocation();
		_preFullscreenWindowSize = getWindowSize();
		ExpediteeApplication.theStage.setFullScreen(true);
	}

	@Override
	public void exitFullscreen()
	{
		ExpediteeApplication.theStage.setFullScreen(false);
		setWindowLocation(_preFullscreenWindowLocation);
		setWindowSize(_preFullscreenWindowSize);
	}

	@Override
	public void pushDrawingSurface(Image image)
	{
		_surfaceStack.push(image);
		_clipStack.push(_currentClip);
		_currentClip = null;
	}

	@Override
	public Image popDrawingSurface()
	{
		// Get the current canvas
		Canvas canvas = _surfaceStack.getCurrentSurface();
		
		// Get its corresponding image
		Image image = _surfaceStack.pop();
		
		// Check we haven't hit the bottom of the stack
		if (image == null) return null;
		
		// Write the canvas contents back to the image
		SnapshotParameters params = new SnapshotParameters();
		params.setFill(Color.TRANSPARENT);
		canvas.snapshot(params, JavaFXMiscManager.getIfUsingJavaFXImageManager().getInternalImage(image));
		
		// Reload the next surface's clip
		_currentClip = _clipStack.pop();
		
		return image;
	}

	@Override
	public void setCompositeAlpha(float alpha)
	{
		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		// Make sure we are setting to the base of the stack
		g.restore();
		
		// Set the blend alpha
		g.setGlobalBlendMode(BlendMode.SRC_OVER);
		g.setGlobalAlpha(alpha);
		
		// Re-clip
		_surfaceStack.ensureClip();
	}

	@Override
	public void setAntialiasing(boolean on)
	{
		// TODO: Implement. cts16
	}
	
	public void setTransientAntialiasingOn() {
		// TODO: Implement. Bryce 
	}
	
	public void setTransientAntialiasingOff() {
		// TODO: Implement. Bryce
	}

	@Override
	public void setFont(Font font)
	{
		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		g.restore();
		
		g.setFont(JavaFXMiscManager.getIfUsingJavaFXFontManager().getInternalFont(font));

		// Re-clip
		_surfaceStack.ensureClip();
	}

	@Override
	public void drawImage(Image image, Point topLeft, Dimension size, double angle, Point cropTopLeft, Dimension cropSize)
	{
		// Can't draw nothing
		if (image == null) return;
		
		// Can't draw nowhere
		if (topLeft == null) return;
		
		// If the cropped area is degenerate, abort
		if (cropSize != null && (cropSize.width <= 0 || cropSize.height <= 0)) return;
		
		// If the crop area is the entire image, pretend we are not cropping
		if (Point.ORIGIN.equals(cropTopLeft) && image.getSize().equals(cropSize)) {
			cropTopLeft = null;
			cropSize = null;
		}

		JavaFXImageManager manager = JavaFXMiscManager.getIfUsingJavaFXImageManager();
		
		// Default is the entire image
		Image croppedImage = image;
		
		// If we are cropping, do this now into a temporary image
		if (cropTopLeft != null && cropSize != null) {
			croppedImage = manager.createImage(cropSize.width, cropSize.height);
			manager.getInternalImage(croppedImage).getPixelWriter().setPixels(0, 0, cropSize.width, cropSize.height, manager.getInternalImage(image).getPixelReader(), cropTopLeft.getX(), cropTopLeft.getY());
		}
		
		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		g.save();
		if (size != null) g.scale(((double) size.width) / croppedImage.getWidth(), ((double) size.height) / croppedImage.getHeight());
		if (angle != 0.0) g.rotate(angle);
		if (topLeft != null) g.translate(topLeft.getX(), topLeft.getY());
		g.drawImage(JavaFXMiscManager.getIfUsingJavaFXImageManager().getInternalImage(croppedImage), 0, 0);
		g.restore();
	}

	@Override
	public void drawRectangle(Point topLeft, Dimension size, double angle, Fill fill, Colour borderColour, Stroke borderStroke, Dimension cornerRadius)
	{
		// Can't draw nowhere
		if (topLeft == null) return;
		
		// Can't draw nothing
		if (size == null) return;
		
		// If not filled and has no border, why bother?
		if (fill == null && (borderColour == null || borderStroke == null)) return;
		
		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		g.save();

		Point centre = new Point(topLeft.getX() + size.width / 2, topLeft.getY() + size.height / 2);
		g.translate(-centre.getX(), -centre.getY());
		if (angle != 0.0) g.rotate(angle);
		g.translate(centre.getX(), centre.getY());
		
		if (fill != null) {
			if (fill instanceof GradientFill) {
				g.setFill(JavaFXConversions.toJavaFXLinearGradient((GradientFill) fill));
			} else {
				g.setFill(JavaFXConversions.toJavaFXColor(fill.getColour()));
			}
			
			if (cornerRadius == null) {
				g.fillRect(topLeft.getX(), topLeft.getY(), size.width, size.height);
			} else {
				g.fillRoundRect(topLeft.getX(), topLeft.getY(), size.width, size.height, cornerRadius.width, cornerRadius.height);
			}
		}
		
		if (borderColour != null && borderStroke != null) {
			g.setStroke(JavaFXConversions.toJavaFXColor(borderColour));
			setStroke(g, borderStroke);

			if (cornerRadius == null) {
				g.strokeRect(topLeft.getX(), topLeft.getY(), size.width, size.height);
			} else {
				g.strokeRoundRect(topLeft.getX(), topLeft.getY(), size.width, size.height, cornerRadius.width, cornerRadius.height);
			}
		}
		
		g.restore();
	}

	@Override
	public void drawOval(Point centre, Dimension diameters, double angle, Fill fill, Colour borderColour, Stroke borderStroke)
	{
		// Can't draw nowhere
		if (centre == null) return;
		
		// Can't draw nothing
		if (diameters == null) return;
		
		// If not filled and has no border, why bother?
		if (fill == null && (borderColour == null || borderStroke == null)) return;
		
		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		g.save();

		if (angle != 0.0) {
			g.translate(-centre.getX(), -centre.getY());
			g.rotate(angle);
			g.translate(centre.getX(), centre.getY());
		}
		
		if (fill != null) {
			if (fill instanceof GradientFill) {
				g.setFill(JavaFXConversions.toJavaFXLinearGradient((GradientFill) fill));
			} else {
				g.setFill(JavaFXConversions.toJavaFXColor(fill.getColour()));
			}
			
			g.fillOval(centre.getX() - diameters.width / 2, centre.getY() - diameters.height / 2, diameters.width, diameters.height);
		}
		
		if (borderColour != null && borderStroke != null) {
			g.setStroke(JavaFXConversions.toJavaFXColor(borderColour));
			setStroke(g, borderStroke);

			g.strokeOval(centre.getX() - diameters.width / 2, centre.getY() - diameters.height / 2, diameters.width, diameters.height);
		}
		
		g.restore();
	}

	@Override
	public void drawPolygon(PolygonBounds points, Point offset, Dimension scale, double angle, Fill fill, Colour borderColour, Stroke borderStroke)
	{
		if (points == null || points.getPointCount() < 2) return;
		if (fill == null && (borderColour == null || borderStroke == null)) return;
		
		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		g.save();

		Point centre = points.getCentre();
		g.translate(-centre.getX(), -centre.getY());
		if (angle != 0.0) g.rotate(angle);
		if (scale != null) {
			double xScale = ((double) scale.width) / (points.getMaxX() - points.getMinX());
			double yScale = ((double) scale.height) / (points.getMaxY() - points.getMinY());
			g.scale(xScale, yScale);
		}
		g.translate(centre.getX(), centre.getY());
		if (offset != null) g.translate(offset.getX(), offset.getY());
		
		Point[] pointArray = points.toArray();
		int nPoints = pointArray.length;
		double[] xArray = new double[nPoints];
		double[] yArray = new double[nPoints];
		
		for (int i = 0; i < nPoints; i++) {
			xArray[i] = pointArray[i].getX() + 0.5;
			yArray[i] = pointArray[i].getY() + 0.5;
		}
		
		if (fill != null) {
			if (fill instanceof GradientFill) {
				g.setFill(JavaFXConversions.toJavaFXLinearGradient((GradientFill) fill));
			} else {
				g.setFill(JavaFXConversions.toJavaFXColor(fill.getColour()));
			}
			
			g.fillPolygon(xArray, yArray, nPoints);
		}
		
		if (borderColour != null && borderStroke != null) {
			g.setStroke(JavaFXConversions.toJavaFXColor(borderColour));
			setStroke(g, borderStroke);

			if (points.isClosed()) {
				g.strokePolygon(xArray, yArray, nPoints);
			} else {
				g.strokePolyline(xArray, yArray, nPoints);
			}
		}
		
		g.restore();
	}

	@Override
	public void drawLine(int x1, int y1, int x2, int y2, Colour colour, Stroke stroke)
	{
		if (colour == null || stroke == null) return;
		
		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		g.save();
		
		g.setStroke(JavaFXConversions.toJavaFXColor(colour));
		setStroke(g, stroke);
		
		g.beginPath();
		g.moveTo(x1 + 0.5, y1 + 0.5);
		g.lineTo(x2 + 0.5, y2 + 0.5);
		g.stroke();
		
		g.restore();
	}

	@Override
	public void drawString(String string, Point position, Font font, Colour colour)
	{
		// TODO: Implement. cts16
	}
	
	@Override
	public void drawTextLayout(TextLayout layout, Point position, Colour colour)
	{
		if (layout == null || position == null || colour == null) return;
		
		JavaFXTextLayoutManager layoutManager = JavaFXMiscManager.getIfUsingJavaFXTextLayoutManager();
		if (layoutManager == null) {
			drawString(layout.getLine(), position, layout.getFont(), colour);
			return;
		}
		
		javafx.scene.text.Text text = layoutManager.getInternalLayout(layout);

		GraphicsContext g = _surfaceStack.getCurrentSurface().getGraphicsContext2D();
		
		g.save();
		
		g.setFont(text.getFont());
		g.setTextAlign(text.getTextAlignment());
		
		g.setFill(JavaFXConversions.toJavaFXColor(colour));
		
		g.fillText(text.getText(), position.getX(), position.getY());
		
		g.restore();
	}

	@Override
	public void setCursor(Cursor cursor)
	{
		// TODO: Implement. cts16
	}

	@Override
	public Dimension getBestCursorSize(Dimension desiredSize)
	{
		// TODO: Implement. cts16
		return new Dimension(32, 32);
	}

	@Override
	public boolean showDialog(String title, String message)
	{
		// TODO: Implement. cts16
		return false;
	}

	@Override
	public Dimension getCurrentDrawingSurfaceSize()
	{
		Canvas currentSurface = _surfaceStack.getCurrentSurface();
		
		int width = (int) currentSurface.getWidth();
		int height = (int) currentSurface.getHeight();
		
		return new Dimension(width, height);
	}
	
	private Point translateCanvasToScreen(Point p)
	{
		if (p == null) return null;
		
		Bounds canvasBounds = ExpediteeApplication.theCanvas.localToScreen(ExpediteeApplication.theCanvas.getBoundsInLocal());
		
		return p.clone().add(canvasBounds.getMinX(), canvasBounds.getMinY());
	}
	
	private Dimension getDecorationDimensions()
	{
		int decorWidth = (int) (ExpediteeApplication.theStage.getWidth() - ExpediteeApplication.theScene.getWidth());
		int decorHeight = (int) (ExpediteeApplication.theStage.getHeight() - ExpediteeApplication.theScene.getHeight());
		
		return new Dimension(decorWidth, decorHeight);
	}

	private void setStroke(GraphicsContext g, Stroke stroke)
	{
		g.setLineWidth(stroke.thickness);
		g.setLineCap(JavaFXConversions.toJavaFXStrokeLineCap(stroke.capType));
		g.setLineJoin(JavaFXConversions.toJavaFXStrokeLineJoin(stroke.joinType));
		g.setLineDashes(Util.toDoubleArray(stroke.dashArray));
		g.setLineDashOffset(stroke.dashPhase);
	}
	
	private Set<Image> createIconSet(Image image)
	{
		if (image == null) return null;
		
		Set<Image> iconSet = new HashSet<Image>();
		
		for (int size = MAX_ICON_SIZE; size >= MIN_ICON_SIZE; size >>= 1) {
			Image tempImage = Image.createImage(size, size);
			this.pushDrawingSurface(tempImage);
			this.drawImage(image, Point.ORIGIN, new Dimension(size, size));
			this.popDrawingSurface();
			iconSet.add(tempImage);
		}
		
		return iconSet;
	}
	
	public static class ExpediteeApplication extends Application {
		
		public static Stage theStage = null;
		public static Scene theScene = null;
		public static Group theGroup = null;
		public static Canvas theCanvas = null;
		
		public static JavaFXGraphicsManager manager = null;

		@Override
		public void start(Stage primaryStage) throws Exception
		{
			// Set up the canvas for drawing operations
			theStage = primaryStage;
			theGroup = new Group();
			theScene = new Scene(theGroup);
			theCanvas = new Canvas();
			theGroup.getChildren().add(theCanvas);
			
			// Set the scene
			manager.setWindowIcon();
			primaryStage.setScene(theScene);
			primaryStage.show();
			
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					synchronized (_jfxAppThread) {
						_jfxAppThread.notify();
					}
				}
			});
		}
		
	}

	@Override
	public int getFontHeight(final Font font) {
		final javafx.scene.text.Font fxFont = JavaFXFontManager.getInstance().getInternalFont(font);
		return (int) fxFont.getSize();
	}

}
